package org.apache.bcel.classfile;

import java.util.function.Predicate;
import java.util.Objects;
import java.util.stream.Stream;
import java.util.Stack;

public class DescendingVisitor implements Visitor
{
    private final JavaClass clazz;
    private final Visitor visitor;
    private final Stack<Object> stack;
    
    public DescendingVisitor(final JavaClass clazz, final Visitor visitor) {
        this.stack = new Stack<Object>();
        this.clazz = clazz;
        this.visitor = visitor;
    }
    
    private <E extends Node> void accept(final E[] node) {
        Stream.of(node).forEach(e -> e.accept(this));
    }
    
    public Object current() {
        return this.stack.peek();
    }
    
    public Object predecessor() {
        return this.predecessor(0);
    }
    
    public Object predecessor(final int level) {
        final int size = this.stack.size();
        if (size < 2 || level < 0) {
            return null;
        }
        return this.stack.elementAt(size - (level + 2));
    }
    
    public void visit() {
        this.clazz.accept(this);
    }
    
    @Override
    public void visitAnnotation(final Annotations annotation) {
        this.stack.push(annotation);
        annotation.accept(this.visitor);
        this.accept(annotation.getAnnotationEntries());
        this.stack.pop();
    }
    
    @Override
    public void visitAnnotationDefault(final AnnotationDefault obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitAnnotationEntry(final AnnotationEntry annotationEntry) {
        this.stack.push(annotationEntry);
        annotationEntry.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitBootstrapMethods(final BootstrapMethods bm) {
        this.stack.push(bm);
        bm.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitCode(final Code code) {
        this.stack.push(code);
        code.accept(this.visitor);
        this.accept(code.getExceptionTable());
        this.accept(code.getAttributes());
        this.stack.pop();
    }
    
    @Override
    public void visitCodeException(final CodeException ce) {
        this.stack.push(ce);
        ce.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantClass(final ConstantClass constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantDouble(final ConstantDouble constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantDynamic(final ConstantDynamic obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantFieldref(final ConstantFieldref constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantFloat(final ConstantFloat constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantInteger(final ConstantInteger constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantInterfaceMethodref(final ConstantInterfaceMethodref constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantInvokeDynamic(final ConstantInvokeDynamic constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantLong(final ConstantLong constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantMethodHandle(final ConstantMethodHandle obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantMethodref(final ConstantMethodref constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantMethodType(final ConstantMethodType obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantModule(final ConstantModule obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantNameAndType(final ConstantNameAndType constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantPackage(final ConstantPackage obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantPool(final ConstantPool cp) {
        this.stack.push(cp);
        cp.accept(this.visitor);
        Stream.of(cp.getConstantPool()).filter(Objects::nonNull).forEach(e -> e.accept(this));
        this.stack.pop();
    }
    
    @Override
    public void visitConstantString(final ConstantString constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantUtf8(final ConstantUtf8 constant) {
        this.stack.push(constant);
        constant.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitConstantValue(final ConstantValue cv) {
        this.stack.push(cv);
        cv.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitDeprecated(final Deprecated attribute) {
        this.stack.push(attribute);
        attribute.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitEnclosingMethod(final EnclosingMethod obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitExceptionTable(final ExceptionTable table) {
        this.stack.push(table);
        table.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitField(final Field field) {
        this.stack.push(field);
        field.accept(this.visitor);
        this.accept(field.getAttributes());
        this.stack.pop();
    }
    
    @Override
    public void visitInnerClass(final InnerClass inner) {
        this.stack.push(inner);
        inner.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitInnerClasses(final InnerClasses ic) {
        this.stack.push(ic);
        ic.accept(this.visitor);
        this.accept(ic.getInnerClasses());
        this.stack.pop();
    }
    
    @Override
    public void visitJavaClass(final JavaClass clazz) {
        this.stack.push(clazz);
        clazz.accept(this.visitor);
        this.accept(clazz.getFields());
        this.accept(clazz.getMethods());
        this.accept(clazz.getAttributes());
        clazz.getConstantPool().accept(this);
        this.stack.pop();
    }
    
    @Override
    public void visitLineNumber(final LineNumber number) {
        this.stack.push(number);
        number.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitLineNumberTable(final LineNumberTable table) {
        this.stack.push(table);
        table.accept(this.visitor);
        this.accept(table.getLineNumberTable());
        this.stack.pop();
    }
    
    @Override
    public void visitLocalVariable(final LocalVariable var) {
        this.stack.push(var);
        var.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitLocalVariableTable(final LocalVariableTable table) {
        this.stack.push(table);
        table.accept(this.visitor);
        this.accept(table.getLocalVariableTable());
        this.stack.pop();
    }
    
    @Override
    public void visitLocalVariableTypeTable(final LocalVariableTypeTable obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitMethod(final Method method) {
        this.stack.push(method);
        method.accept(this.visitor);
        this.accept(method.getAttributes());
        this.stack.pop();
    }
    
    @Override
    public void visitMethodParameter(final MethodParameter obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitMethodParameters(final MethodParameters obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        Stream.of(obj.getParameters()).forEach(e -> e.accept(this));
        this.stack.pop();
    }
    
    @Override
    public void visitModule(final Module obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.accept(obj.getRequiresTable());
        this.accept(obj.getExportsTable());
        this.accept(obj.getOpensTable());
        this.accept(obj.getProvidesTable());
        this.stack.pop();
    }
    
    @Override
    public void visitModuleExports(final ModuleExports obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitModuleMainClass(final ModuleMainClass obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitModuleOpens(final ModuleOpens obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitModulePackages(final ModulePackages obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitModuleProvides(final ModuleProvides obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitModuleRequires(final ModuleRequires obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitNestHost(final NestHost obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitNestMembers(final NestMembers obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitParameterAnnotation(final ParameterAnnotations obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitParameterAnnotationEntry(final ParameterAnnotationEntry obj) {
        this.stack.push(obj);
        obj.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitSignature(final Signature attribute) {
        this.stack.push(attribute);
        attribute.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitSourceFile(final SourceFile attribute) {
        this.stack.push(attribute);
        attribute.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitStackMap(final StackMap table) {
        this.stack.push(table);
        table.accept(this.visitor);
        this.accept(table.getStackMap());
        this.stack.pop();
    }
    
    @Override
    public void visitStackMapEntry(final StackMapEntry var) {
        this.stack.push(var);
        var.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitSynthetic(final Synthetic attribute) {
        this.stack.push(attribute);
        attribute.accept(this.visitor);
        this.stack.pop();
    }
    
    @Override
    public void visitUnknown(final Unknown attribute) {
        this.stack.push(attribute);
        attribute.accept(this.visitor);
        this.stack.pop();
    }
}
